/* src/net/server.c 
 * ============================================================
 * OCSP Responder
 * (c) 2001 by Massimiliano Pala
 * OpenCA released software
 * ============================================================
 */
 
#include <openca/general.h>

#include <unistd.h>
#include <strings.h>
#include <resolv.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <openssl/conf.h>
#include <openssl/txt_db.h>

#include "../tools/support.h"
#include "../crypto/ocsp_response.h"

#define MAX_IN_LINES	16536

/* Custom structures */
typedef struct http_request
	{
		int method;
		long header_size;
		char *target;
		char *content_type;
		long data_size;
		char *data;
	} HTTP_REQUEST;

typedef struct in_lines
	{
		long lines;
		char **ln_pnt;
	} IN_LINES;

/* Functions prototypes */
char *get_http_header ( char *tmp_text, int recv_size );
HTTP_REQUEST *parse_http_request ( char *req, long size );
IN_LINES *new_in_lines( char *txt );
int send_ocsp_response(BIO *cbio, OCSP_RESPONSE *resp);

/* External General Variables */
extern int debug;
extern int verbose;

int start_server( char *bind_s, char *port_s, int queue, long max_size, CONF *conf,
	          EVP_PKEY *pkey, X509 *ocspd_cert, X509 *cacert, TXT_DB *db ) {

	BIO *bio_socket, *tmp;
	BIO *bio_mem = NULL;
	BIO *out = NULL;
	HTTP_REQUEST *http_parsed_req = NULL;
	OCSP_REQUEST *OCSP_req = NULL;
	OCSP_RESPONSE *OCSP_resp = NULL;

	char *buf = NULL;
	char tmp_socket[1024];
	char *addr = NULL;

	int i = 0;
	long buf_len = 0;

	/* Let's add for later SSL support */
	SSL_load_error_strings();
	OpenSSL_add_ssl_algorithms();

	/* No max size under 1024 */
	if( max_size < 1024 ) max_size = 1024;
	
	if((addr = (char *) OPENSSL_malloc (strlen(bind_s)+strlen(port_s)+3))
								== NULL ) {
		syslog(LOG_ERR, "Cannot alloc addr");
		exit( -5 );
	}

	/* We build the "host:port" string to be used */
	sprintf( addr, "%s:%s", bind_s, port_s );

	if((bio_socket = BIO_new_accept( addr )) == NULL ) {
		syslog(LOG_ERR, "Cannot bind to %s", addr);
		exit(-10);
	}
	if (verbose)
		syslog(LOG_INFO, "successfully binded to %s", addr);

	buf = (char *) OPENSSL_malloc ( max_size );
	while ( 1 ) {

		if( BIO_do_accept(bio_socket) <= 0) {
			syslog(LOG_ERR, "Cannot accept (bio err on %s)", addr);
			goto sock_err;
		}

		memset( buf, 0, max_size );
		buf_len = 0;
		for(;;) {
			memset( tmp_socket, 0, 1024 );
			i=BIO_read( bio_socket, tmp_socket, 1024);

			/* Check for error or stream end */
			if( i < 0 ) {
				syslog( LOG_ERR, "Error in data receiving, dropping socket" );
				goto sock_err;
			} else if ( i == 0 ) {
				break;
			};

			/* Check we do not receive too much data, this could be 
			 * removed later using the double BIO ends for request
			 * parsing 
			 */
			if( max_size <= buf_len + i ) {
				syslog( LOG_ERR, "Request too long, rejecting");
				goto sock_err;
			}

			memcpy( buf + buf_len, tmp_socket, i );
			buf_len += i;
			syslog( LOG_ERR, "read %i ( total buf_len %ld)", i, buf_len);

			if( i < 1024 ) break;

		}
		syslog( LOG_ERR, "HTTP Request total size %ld bytes", buf_len);

		/* Process the request */
		if( (http_parsed_req = parse_http_request( buf, buf_len )) == NULL ) {
			syslog(LOG_ERR,"Cannot parse http request, rejecting");
			goto sock_err;
		}

		if( (bio_mem = BIO_new( BIO_s_mem() )) == NULL ) {
			syslog( LOG_ERR, "Canno open a mem bio!");
			goto sock_err;
		}

		i=BIO_write( bio_mem, http_parsed_req->data, http_parsed_req->data_size );

		/* Read OCSP request (if any) */
		if( (OCSP_req = d2i_OCSP_REQUEST_bio(bio_mem, NULL)) == NULL ) {
			syslog( LOG_ERR, "Cannot read request");
			goto sock_err;
		}

		i = make_ocsp_response( &OCSP_resp, OCSP_req, db, cacert, 
				ocspd_cert, pkey, NULL, OCSP_NOCERTS, 0, -1 );

		/* Send the response */
		send_ocsp_response( bio_socket, OCSP_resp );


		out = BIO_new_fp(stdout, BIO_NOCLOSE);
		OCSP_REQUEST_print( out, OCSP_req, 0);
		OCSP_RESPONSE_print( out, OCSP_resp, 0);

		send_ocsp_response( out, OCSP_resp );


sock_err:
		/* Close the socket */
		if( bio_mem ) BIO_free (bio_mem);
		if( out ) BIO_free (out);
		tmp = BIO_pop(bio_socket);
		BIO_free_all(tmp);

	}

	if( buf ) OPENSSL_free (buf);

	return(0);

}

char *get_http_header ( char *tmp_text, int recv_size ) {

	char *head_end = "\r\n\r\n";
	char *ret = NULL;
	char *tmp = NULL;
	long mylen = 0;

	if( (tmp = strstr( tmp_text, head_end )) == NULL ) {
		return(NULL);
	}

	mylen = (long) (long)tmp - (long)tmp_text;
	if ( (mylen > recv_size) || (mylen == 0) )
			return(NULL);

	ret = (char *) OPENSSL_malloc ( mylen + 1 );
	memcpy(ret, tmp_text, mylen);
	ret[mylen] = 0;

	printf( ret );
	return (ret);
}

HTTP_REQUEST *parse_http_request ( char *req, long size ) {

	HTTP_REQUEST *http_req = NULL;
	IN_LINES *lines;
	char *header = NULL;
	long i;
	

	if(! req ) return(NULL);

	if( (http_req = (HTTP_REQUEST *) OPENSSL_malloc (sizeof(HTTP_REQUEST))) == NULL)
			return(NULL);

	syslog( LOG_INFO,"http_req allocated" );

	header = get_http_header( req, size );
	lines = new_in_lines( header );

	http_req->header_size = strlen(header) + 2;
	http_req->data = req + strlen( header ) + 4;

	if( lines->lines < 0 ) return (NULL);

	if( strncmp_nocase( lines->ln_pnt[0], "POST", 4) == 0)
		{
		http_req->method = HTTP_POST;
	       	}
       	else if ( strncmp_nocase( lines->ln_pnt[0], "GET", 3) == 0)
       		{
		http_req->method = HTTP_GET;
		}
       	else 
		{
		syslog( LOG_ERR,"HTTP Method unsupported." );
		OPENSSL_free( http_req );
		return(NULL);
		}

	for ( i = 1; i <= lines->lines; i++ ) {

		syslog( LOG_INFO,"%ld -> %s", i, lines->ln_pnt[i] );
		if( strncmp_nocase( lines->ln_pnt[i], "Content-Type: ", 14 ) == 0 ) {
			if ( strncmp_nocase ( lines->ln_pnt[i] + 14, OCSP_REQ_TYPE, strlen(OCSP_REQ_TYPE))) {
				syslog( LOG_ERR, "Client sent wrong content type %s",
					lines->ln_pnt[i] + 14 );
				goto err;
			}
			http_req->method = HTTP_POST;
		}

		if( strncmp_nocase( lines->ln_pnt[i], "Content-Length: ", 16 ) == 0 )
			http_req->data_size = atol( lines->ln_pnt[i] + 16 );

		if( strncmp_nocase( lines->ln_pnt[i], "Host: ", 6 ) == 0 )
			http_req->target = lines->ln_pnt[i] + 6;
	}

	return(http_req);

err:
	if( http_req ) OPENSSL_free (http_req);
	return (NULL);
}

IN_LINES *new_in_lines( char *txt ) {
	
	IN_LINES *in = NULL;
	char *p1 = NULL;
	char *p2 = NULL;
	char *tmp = NULL;
	int cont = 1;
	long mylen;

	if((in = (IN_LINES *) OPENSSL_malloc ( sizeof( IN_LINES ))) == NULL ) {
		/* Memory allocation error */
		exit(1);
	}

	in->lines = -1;

	if((in->ln_pnt = (char **) OPENSSL_malloc ( sizeof(char *) * MAX_IN_LINES )) == NULL ) {
		/* Memory allocation error */
		exit(1);
	}

	p1 = txt;
	while ( (cont)  && (in->lines < MAX_IN_LINES) ) {

		if( (p2 = strstr( p1, "\r\n" )) == NULL ) {
			mylen = strlen(p1);
			p2 = p1 + mylen - 1;
			cont = 0;
		} else {
			mylen = (long) p2 - (long) p1;
		}

		in->lines++;

		tmp = (char *) OPENSSL_malloc ( mylen + 1 );
		strncpy( tmp, p1, mylen );
		tmp[mylen] = 0;

		in->ln_pnt[in->lines] = tmp;

		p1 = p2 + 2;
	}

	return(in);
}

int send_ocsp_response(BIO *cbio, OCSP_RESPONSE *resp) {
	char http_resp[] = "HTTP/1.0 200 OK\r\nContent-type: application/ocsp-response\r\n"
			   "Content-Length: %d\r\n\r\n";

	if (!cbio)
		return 0;

	BIO_printf(cbio, http_resp, i2d_OCSP_RESPONSE(resp, NULL));
	i2d_OCSP_RESPONSE_bio(cbio, resp);
	BIO_flush(cbio);
	return 1;
}

